from redis import StrictRedis


redis = StrictRedis(
    host='localhost',
    port='7001',
    db='0',
)


def get_key_weekly_price_pct_chg_sd(symbol: str) -> str:
    return f'weekly_price_pct_chg_sd_{symbol}'



class OptionSellOpportunity(object):

    def __init__(
        self,
        symbol: str,
        last_price: float,
        pct_chg_so_far: float,
        sd_this_week: float,
        mean_this_week: float,
        sd_next_week: float,
        mean_next_week: float,
    ) -> None:
        self._symbol = symbol
        self._last_price = last_price
        self._pct_chg_so_far = pct_chg_so_far
        self._sd_this_week = sd_this_week
        self._mean_this_week = mean_this_week
        self._sd_next_week = sd_next_week
        self._mean_next_week = mean_next_week

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def last_price(self) -> float:
        return self._last_price

    @property
    def pct_chg_so_far(self) -> float:
        return self._pct_chg_so_far

    @property
    def sd_this_week(self) -> float:
        return self._sd_this_week

    @property
    def mean_this_week(self) -> float:
        return self._mean_this_week

    @property
    def sd_next_week(self) -> float:
        return self._sd_next_week

    @property
    def mean_next_week(self) -> float:
        return self._mean_next_week

    @property
    def z_score_this_week(self) -> float:
        return abs(self.pct_chg_so_far - self.mean_this_week) / self.sd_this_week

    @property
    def z_score_next_week(self) -> float:
        if self.sd_next_week is None:
            return 0.0
        else:
            return abs(self.pct_chg_so_far - self.mean_next_week) / self.sd_next_week



from copy import copy
from typing import Dict
from typing import List
from typing import Optional

from report.symbolhistory.symbolhistorycontextholding import SymbolHistoryContextHolding
from report.symbolhistory.symbolhistorycontextsnapshot import SymbolHistoryContextSnapshot
from report.symbolhistory.symbolhistoryevent import SymbolHistoryEvent
from robinhood.lib.marketplace.optioninstrument import OptionInstrument


class SymbolHistoryContext(object):

    def __init__(self, symbol: str) -> None:
        self._profit = 0.0
        self._stock_holding = SymbolHistoryContextHolding(
            symbol=symbol,
            quantity=0,
            avg_unit_price=0.0,
            option_instrument=None,
        )
        self._option_sell_holding: Dict[str, SymbolHistoryContextHolding] = {}
        self._option_buy_holding: Dict[str, SymbolHistoryContextHolding] = {}
        self._tmp_call_assignment: List[SymbolHistoryContextHolding] = []
        self._tmp_stock_sold: Optional[SymbolHistoryContextHolding] = None

    @property
    def stock_cost_basis(self) -> float:
        return self._stock_holding.quantity * self._stock_holding.avg_unit_price

    @property
    def option_sell_holding(self) -> Dict[str, SymbolHistoryContextHolding]:
        return self._option_sell_holding

    @property
    def option_buy_holding(self) -> Dict[str, SymbolHistoryContextHolding]:
        return self._option_buy_holding

    def take_snapshot(self) -> SymbolHistoryContextSnapshot:
        return SymbolHistoryContextSnapshot(
            profit=self._profit,
            stock_holding=self._stock_holding,
            option_sell_holding=copy(self._option_sell_holding),
            option_buy_holding=copy(self._option_buy_holding),
        )

    def limit_or_market_buy_stock(self, quantity: int, price: float) -> None:
        total_cost = self.stock_cost_basis + quantity * price
        total_quantity = self._stock_holding.quantity + quantity
        self._stock_holding = SymbolHistoryContextHolding(
            symbol=self._stock_holding.symbol,
            quantity=self._stock_holding.quantity + quantity,
            avg_unit_price=total_cost / total_quantity,
            option_instrument=None,
        )

    def limit_or_market_sell_stock(self, quantity: int, price: float) -> None:
        self._profit += (price - self._stock_holding.avg_unit_price) * quantity
        self._stock_holding = SymbolHistoryContextHolding(
            symbol=self._stock_holding.symbol,
            quantity=self._stock_holding.quantity - quantity,
            avg_unit_price=self._stock_holding.avg_unit_price,
            option_instrument=None,
        )

    def sell_to_open(
        self,
        symbol: str,
        quantity: int,
        unit_price: float,
        option_instrument: OptionInstrument,
    ) -> None:
        self._buy_or_sell_to_open(
            buy_or_sell='sell',
            symbol=symbol,
            quantity=quantity,
            unit_price=unit_price,
            option_instrument=option_instrument,
        )

    def buy_to_close(
        self,
        symbol: str,
        quantity: int,
        unit_price: float,
    ) -> None:
        self._buy_or_sell_to_close(
            buy_or_sell='buy',
            symbol=symbol,
            quantity=quantity,
            unit_price=unit_price,
        )

    def buy_to_open(
        self,
        symbol: str,
        quantity: int,
        unit_price: float,
        option_instrument: OptionInstrument,
    ) -> None:
        self._buy_or_sell_to_open(
            buy_or_sell='buy',
            symbol=symbol,
            quantity=quantity,
            unit_price=unit_price,
            option_instrument=option_instrument,
        )

    def sell_to_close(
        self,
        symbol: str,
        quantity: int,
        unit_price: float,
    ) -> None:
        self._buy_or_sell_to_close(
            buy_or_sell='sell',
            symbol=symbol,
            quantity=quantity,
            unit_price=unit_price,
        )

    def call_assignment(self, symbol: str, quantity: int, price: float, event: SymbolHistoryEvent) -> None:
        holding = self._option_sell_holding[symbol]
        current_avg_unit_price = holding.avg_unit_price
        new_quantity = holding.quantity - quantity
        if new_quantity == 0:
            del self._option_sell_holding[symbol]
        else:
            self._option_sell_holding[symbol] = SymbolHistoryContextHolding(
                symbol=symbol,
                quantity=new_quantity,
                avg_unit_price=holding.avg_unit_price,
                option_instrument=holding.option_instrument,
            )

        shares_sold = quantity * 100
        if self._stock_holding.quantity - shares_sold < 0:
            new_unit_price = price + current_avg_unit_price
            self._tmp_call_assignment.append(
                SymbolHistoryContextHolding(
                    symbol=symbol,
                    quantity=quantity,
                    avg_unit_price=new_unit_price,
                    option_instrument=None,
                ),
            )
            event.unit_price = new_unit_price
            return

        profit_per_share = price - self._stock_holding.avg_unit_price + current_avg_unit_price
        self._profit += profit_per_share * shares_sold
        self._stock_holding = SymbolHistoryContextHolding(
            symbol=self._stock_holding.symbol,
            quantity=self._stock_holding.quantity - shares_sold,
            avg_unit_price=self._stock_holding.avg_unit_price,
            option_instrument=None,
        )

    def put_assignment(
        self,
        symbol: str,
        num_of_contracts: int,
        strike_price: float,
    ) -> None:
        """
        This function makes changes to the portfolio when a put assignment event
        happens.

        Arguments:
            symbol: the symbol of the option instrument being assigned.
            num_of_contracts: the number of contracts being assigned.
            strike_price: the strike price of the contracts being assigned.
        """
        holding = self._option_sell_holding[symbol]
        new_quantity = holding.quantity - num_of_contracts
        if new_quantity == 0:
            del self._option_sell_holding[symbol]
        else:
            self._option_sell_holding[symbol] = SymbolHistoryContextHolding(
                symbol=symbol,
                quantity=new_quantity,
                avg_unit_price=holding.avg_unit_price,
                option_instrument=holding.option_instrument,
            )

        shares_bought = num_of_contracts * 100
        shares_avg_unit_cost = strike_price - holding.avg_unit_price
        shares_cost_basis = shares_avg_unit_cost * shares_bought
        if self._tmp_stock_sold is None:
            new_shares_holding = self._stock_holding.quantity + shares_bought
            avg_unit_price = (
                (self.stock_cost_basis + shares_cost_basis)
                / (self._stock_holding.quantity + shares_bought)
            )
        elif self._tmp_stock_sold.quantity >= shares_bought:
            new_shares_holding = self._stock_holding.quantity
            avg_unit_price = self._stock_holding.avg_unit_price
            profit = (self._tmp_stock_sold.avg_unit_price - shares_avg_unit_cost) * shares_bought
            self._profit += profit
            self._tmp_stock_sold = SymbolHistoryContextHolding(
                symbol=self._stock_holding.symbol,
                quantity=self._tmp_stock_sold.quantity - shares_bought,
                avg_unit_price=self._tmp_stock_sold.avg_unit_price,
                option_instrument=None,
            )
            if self._tmp_stock_sold.quantity == 0:
                self._tmp_stock_sold = None
        else:
            assert self._tmp_stock_sold.quantity < shares_bought
            new_shares_holding = shares_bought - self._tmp_stock_sold.quantity
            avg_unit_price = shares_avg_unit_cost
            profit = (self._tmp_stock_sold.avg_unit_price - shares_avg_unit_cost) * self._tmp_stock_sold.quantity
            self._profit += profit
            self._tmp_stock_sold = None

        self._stock_holding = SymbolHistoryContextHolding(
            symbol=self._stock_holding.symbol,
            quantity=new_shares_holding,
            avg_unit_price=avg_unit_price,
            option_instrument=None,
        )

    def put_exercise(
        self,
        symbol: str,
        num_of_contracts: int,
        avg_unit_price: float,
        strike_price: float,
    ) -> None:
        """
        This function makes changes to the portfolio when a put exercise event
        happens.

        Arguments:
            symbol: the symbol of the option instrument being exercised.
            num_of_contracts: the number of contracts being exercised.
            avg_unit_price: the average unit price of the contracts being exercised.
            strike_price: the strike price of the contracts being exercised.
        """
        shares_sold = num_of_contracts * 100
        share_sold_price = strike_price - avg_unit_price
        if self._stock_holding.quantity >= shares_sold:
            profit = (self._stock_holding.avg_unit_price - share_sold_price) * shares_sold
            self._profit += profit
            self._stock_holding = SymbolHistoryContextHolding(
                symbol=self._stock_holding.symbol,
                quantity=self._stock_holding.quantity - shares_sold,
                avg_unit_price=self._stock_holding.avg_unit_price,
                option_instrument=None,
            )
        else:
            cost_basis = shares_sold * share_sold_price

            if self._tmp_stock_sold is None:
                existing_quantity = 0
            else:
                existing_quantity = self._tmp_stock_sold.quantity
                cost_basis += self._tmp_stock_sold.avg_unit_price * existing_quantity

            self._tmp_stock_sold = SymbolHistoryContextHolding(
                symbol=self._stock_holding.symbol,
                quantity=existing_quantity + shares_sold,
                avg_unit_price=cost_basis / shares_sold,
                option_instrument=None,
            )

        holding = self._option_buy_holding[symbol]
        new_quantity = holding.quantity - num_of_contracts
        if new_quantity == 0:
            del self._option_buy_holding[symbol]
        else:
            self._option_buy_holding[symbol] = SymbolHistoryContextHolding(
                symbol=symbol,
                quantity=new_quantity,
                avg_unit_price=holding.avg_unit_price,
                option_instrument=holding.option_instrument,
            )

    def _call_exercise_for_assignment(
        self,
        symbol: str,
        quantity: int,
        price: float,
        strike_price: float,
    ) -> bool:
        for holding in self._tmp_call_assignment:
            if holding.quantity == quantity:
                profit = (holding.avg_unit_price - strike_price - price) * quantity * 100
                self._profit += profit
                del holding
                return True

        return False

    def call_exercise(
        self,
        symbol: str,
        quantity: int,
        price: float,
        strike_price: float,
    ) -> None:
        if not self._call_exercise_for_assignment(
            symbol=symbol,
            quantity=quantity,
            price=price,
            strike_price=strike_price,
        ):
            shares_bought = quantity * 100
            effective_cost_per_share = strike_price + price
            new_cost = effective_cost_per_share * shares_bought
            total_cost = self.stock_cost_basis + new_cost
            new_quantity = self._stock_holding.quantity + shares_bought
            avg_unit_price = 0.0 if new_quantity == 0.0 else total_cost / new_quantity

            self._stock_holding = SymbolHistoryContextHolding(
                symbol=self._stock_holding.symbol,
                quantity=new_quantity,
                avg_unit_price=avg_unit_price,
                option_instrument=None,
            )

        holding = self._option_buy_holding[symbol]
        new_quantity = holding.quantity - quantity
        if new_quantity == 0:
            del self._option_buy_holding[symbol]
        else:
            self._option_buy_holding[symbol] = SymbolHistoryContextHolding(
                symbol=symbol,
                quantity=new_quantity,
                avg_unit_price=holding.avg_unit_price,
                option_instrument=holding.option_instrument,
            )

    def option_expire(self, symbol: str, quantity: int) -> None:
        if symbol in self._option_buy_holding:
            holding_dict = self._option_buy_holding
            holding = self._option_buy_holding[symbol]
            profit = -holding.quantity * holding.avg_unit_price * 100
        elif symbol in self._option_sell_holding:
            holding_dict = self._option_sell_holding
            holding = self._option_sell_holding[symbol]
            profit = holding.quantity * holding.avg_unit_price * 100
        else:
            return

        assert holding.quantity - quantity == 0
        del holding_dict[symbol]
        self._profit += profit

    def _buy_or_sell_to_open(
        self,
        buy_or_sell: str,
        symbol: str,
        quantity: int,
        unit_price: float,
        option_instrument: OptionInstrument,
    ) -> None:
        if buy_or_sell == 'buy':
            option_holding_dict = self._option_buy_holding
        else:
            option_holding_dict = self._option_sell_holding

        if symbol in option_holding_dict:
            holding = option_holding_dict[symbol]
            new_quantity = holding.quantity + quantity
            new_price = (holding.quantity * holding.avg_unit_price + quantity * unit_price) / new_quantity
        else:
            new_quantity = quantity
            new_price = unit_price

        option_holding_dict[symbol] = SymbolHistoryContextHolding(
            symbol=symbol,
            quantity=new_quantity,
            avg_unit_price=new_price,
            option_instrument=option_instrument,
        )

    def _buy_or_sell_to_close(
        self,
        buy_or_sell: str,
        symbol: str,
        quantity: int,
        unit_price: float,
    ) -> None:
        if buy_or_sell == 'buy':
            option_holding_dict = self._option_sell_holding
            unit_profit = option_holding_dict[symbol].avg_unit_price - unit_price
        else:
            option_holding_dict = self._option_buy_holding
            unit_profit = unit_price - option_holding_dict[symbol].avg_unit_price

        new_quantity = option_holding_dict[symbol].quantity - quantity
        self._profit += quantity * unit_profit * 100
        if new_quantity == 0:
            del option_holding_dict[symbol]
        else:
            option_holding_dict[symbol] = SymbolHistoryContextHolding(
                symbol=symbol,
                quantity=new_quantity,
                avg_unit_price=option_holding_dict[symbol].avg_unit_price,
                option_instrument=option_holding_dict[symbol].option_instrument,
            )



import json
from datetime import datetime
from time import sleep
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from pandas import DataFrame
from pandas import Series

from alphavantage.alphavantageclient import AlphaVantageClient
from alphavantage.pricehistory.pricehistory import PriceHistory
from data.redis import get_key_weekly_price_pct_chg_sd
from data.redis import redis
from data.watchlist import ALL


class CalcStockWeeklyPctChgSd(object):

    def __init__(self) -> None:
        self._alphavantage = AlphaVantageClient()

    def run(self) -> None:
        for symbol in ALL:
            if symbol in {'RHT'}:
                continue

            print(symbol)
            weekly_pct_chg_sd = self._get_weekly_pct_chg_sd(symbol=symbol)
            if weekly_pct_chg_sd is None:
                continue

            redis.set(
                name=get_key_weekly_price_pct_chg_sd(symbol=symbol),
                value=json.dumps(weekly_pct_chg_sd),
            )

    def _get_price_history(self, symbol: str) -> Optional[PriceHistory]:
        try:
            price_history = self._alphavantage.get_price_history(
                symbol=symbol,
                time_window='daily',
                output_size='full',
            )

            return price_history
        except KeyError:
            sleep(5)
            return self._get_price_history(symbol=symbol)

    def _get_time_series_df(self, symbol: str) -> Optional[DataFrame]:
        price_history = self._get_price_history(symbol=symbol)
        if price_history is None:
            return None

        data: List[Dict[str, Any]] = []
        for ds, item in price_history.time_series.items():
            row: Dict[str, Any] = {
                'ds': ds,
                'open_price': item.adjusted_open_price,
                'close_price': item.adjusted_close_price,
            }
            data.append(row)

        return DataFrame(data)

    def _enrich_time_series_df(self, df: DataFrame) -> DataFrame:
        df['dow'] = df.ds.map(self._get_dow)
        df['woy'] = df.ds.map(self._get_woy)
        df['woy_full'] = df.apply(self._get_woy_full, axis=1)
        df['next_woy_full'] = df.apply(self._get_next_woy_full, axis=1)
        return df

    def _get_last_trading_day_df(self, df: DataFrame) -> DataFrame:
        idx = df.dow == df.groupby('woy_full').dow.transform(max)
        last_trading_day_df = df[idx][['close_price', 'woy_full']].set_index('woy_full')

        return last_trading_day_df

    def _get_first_trading_day_df(self, df: DataFrame) -> DataFrame:
        idx = df.dow == df.groupby('woy_full').dow.transform(min)
        first_trading_day_df = df[idx][['ds', 'open_price', 'woy_full']].set_index('woy_full')

        return first_trading_day_df

    def _get_weekly_pct_chg_sd(self, symbol: str) -> Optional[Dict[str, Any]]:
        df = self._get_time_series_df(symbol=symbol)
        if df is None:
            return None

        if df.shape[0] >= 360:
            n = 180
        elif df.shape[0] >= 180:
            n = 90
        else:
            return None

        enriched_df = self._enrich_time_series_df(df=df)
        last_trading_day_df = self._get_last_trading_day_df(df=df)

        joined = enriched_df.join(
            other=last_trading_day_df,
            on='woy_full',
            rsuffix='_this_week',
        ).join(
            other=last_trading_day_df,
            on='next_woy_full',
            rsuffix='_next_week',
        )

        joined['to_weekend_pct_chg'] = joined.close_price_this_week / joined.open_price - 1
        joined['to_next_weekend_pct_chg'] = joined.close_price_next_week / joined.open_price - 1

        joined['rolling_nd_to_weekend_pct_chg_std'] = joined.to_weekend_pct_chg.rolling(window=n).std()
        joined['rolling_nd_to_weekend_pct_chg_mean'] = joined.to_weekend_pct_chg.rolling(window=n).mean()
        joined['rolling_nd_to_next_weekend_pct_chg_std'] = joined.to_next_weekend_pct_chg.rolling(window=n).std()
        joined['rolling_nd_to_next_weekend_pct_chg_mean'] = joined.to_next_weekend_pct_chg.rolling(window=n).mean()

        print(joined['rolling_nd_to_next_weekend_pct_chg_std'])
        print(joined['rolling_nd_to_next_weekend_pct_chg_mean'])
        print(n)
        print(joined.tail(60))

        this_week = joined.iloc[-1].rolling_nd_to_weekend_pct_chg_std
        this_week_mean = joined.iloc[-1].rolling_nd_to_weekend_pct_chg_mean
        this_week_ds = joined.iloc[-1].ds
        next_week_idx = joined.rolling_nd_to_next_weekend_pct_chg_std.last_valid_index()
        if next_week_idx is None:
            next_week = None
            next_week_mean = None
            next_week_ds = None
        else:
            next_week = joined.iloc[next_week_idx].rolling_nd_to_next_weekend_pct_chg_std
            next_week_mean = joined.iloc[next_week_idx].rolling_nd_to_next_weekend_pct_chg_mean
            next_week_ds = joined.iloc[next_week_idx].ds

        first_trading_day_df = self._get_first_trading_day_df(df=df)
        first_trading_day_ds = first_trading_day_df.iloc[-1].ds
        first_trading_day_open_price = first_trading_day_df.iloc[-1].open_price

        return {
            'this_week': this_week,
            'this_week_mean': this_week_mean,
            'this_week_ds': this_week_ds,
            'next_week': next_week,
            'next_week_mean': next_week_mean,
            'next_week_ds': next_week_ds,
            'first_trading_day_open_price': first_trading_day_open_price,
            'first_trading_day_ds': first_trading_day_ds,
        }

    def _get_dow(self, ds: str) -> int:
        return datetime.strptime(ds, '%Y-%m-%d').weekday()

    def _get_woy(self, ds: str) -> int:
        return datetime.strptime(ds, '%Y-%m-%d').isocalendar()[1]

    def _get_woy_full(self, row: Series) -> str:
        year = row.ds[:4]
        woy = str(row.woy)
        return f'{year}/{woy}'

    def _get_next_woy_full(self, row: Series) -> str:
        new_woy = (row.woy + 1) % 52
        if new_woy == 0:
            new_woy = 52

        new_year = int(row.ds[:4])
        if new_woy == 1:
            new_year += 1

        return f'{new_year}/{new_woy}'



from collections import OrderedDict
from typing import Any

from asciitree import LeftAligned

from report.symbolhistory.symbolhistory import SymbolHistory
from robinhood.lib.account.account import account
from robinhood.lib.helpers import format_price
from scripts.libs.slack.slackclient import slack_client


class EarningReport(object):

    def __init__(self) -> None:
        pass

    def run(self) -> None:
        report_txt = self._gen_earning_report()
        slack_client.post_to_earning(msg=f'```{report_txt}```')

    def _gen_earning_report(self) -> str:
        account.update_in_parallel()
        symbol_history = SymbolHistory(
            stock_orders=account.stock_orders,
            option_orders=account.option_orders,
            option_events=account.option_events,
        )
        (
            _,
            total_profit,
            total_profit_by_exp_date,
            total_profit_by_exp_date_by_symbol,
        ) = symbol_history.get_earning_data()

        total_profit_key = f'Total Profit: {format_price(total_profit)}'
        profit_by_exp_date_tree: OrderedDict[str, Any] = OrderedDict()
        for exp_date in sorted(total_profit_by_exp_date.keys()):
            profit = total_profit_by_exp_date[exp_date]
            symbol_profit = [
                (symbol, symbol_profit)
                for symbol, symbol_profit in total_profit_by_exp_date_by_symbol[exp_date].items()
            ]
            profit_by_exp_date_tree[f'By {exp_date}: {format_price(profit)}'] = {
                '{:5s} ${:.2f}'.format(symbol, symbol_profit): {}
                for symbol, symbol_profit in sorted(symbol_profit, key=lambda x: x[1])
            }
        profit_tree = {total_profit_key: profit_by_exp_date_tree}
        tree = LeftAligned()

        return str(tree(profit_tree))



import json
import time
from os import environ
from typing import Dict

import requests


class SlackClient(object):
    EARNING_ID = 'CQAJU08DR'
    OPTION_BOT_ID = 'CQPRPNC7R'

    def __init__(self) -> None:
        self._token = environ.get('SLACK_TOKEN')
        if self._token is None:
            raise Exception('Missing SLACK_TOKEN')

    def _gen_header(self) -> Dict[str, str]:
        return {
            'Authorization': f'Bearer {self._token}',
            'Content-Type': 'application/json',
        }

    def _post_to_channel(self, msg: str, channel_id: str) -> None:
        session = requests.session()
        session.headers = self._gen_header()
        data = str.encode(json.dumps({
            'channel': channel_id,
            'text': msg,
        }))
        response = session.post('https://slack.com/api/chat.postMessage', data=data)
        if response.status_code != 200:
            time.sleep(5)
            self._post_to_channel(msg=msg, channel_id=channel_id)

    def post_to_option_bot(self, msg: str) -> None:
        self._post_to_channel(msg=msg, channel_id=self.OPTION_BOT_ID)

    def post_to_earning(self, msg: str) -> None:
        self._post_to_channel(msg=msg, channel_id=self.EARNING_ID)


slack_client = SlackClient()



import click


@click.command()
def calc_stock_weekly_pct_chg_sd() -> None:
    from scripts.calcstockweeklypctchgsd import CalcStockWeeklyPctChgSd
    CalcStockWeeklyPctChgSd().run()


@click.command()
def run_option_sell_screener() -> None:
    from scripts.optionsellscreener import OptionSellScreener
    OptionSellScreener().run()


@click.command()
def slack_earning_report() -> None:
    from scripts.earningreport import EarningReport
    EarningReport().run()


@click.group()
def manage() -> None:
    pass


manage.add_command(calc_stock_weekly_pct_chg_sd)
manage.add_command(run_option_sell_screener)
manage.add_command(slack_earning_report)

if __name__ == '__main__':
    manage()



import json
from datetime import datetime
from typing import Dict
from typing import Optional

from terminaltables import AsciiTable

from data.redis import get_key_weekly_price_pct_chg_sd
from data.redis import redis
from data.watchlist import ALL
from robinhood.lib.marketplace.equity import Equity
from robinhood.lib.marketplace.marketplace import marketplace
from scripts.libs.optionsellopportunity import OptionSellOpportunity
from scripts.libs.slack.slackclient import slack_client


class OptionSellScreener(object):

    def __init__(self) -> None:
        pass

    def run(self) -> None:
        equities = marketplace.get_equities(symbols={symbol for symbol in ALL})
        opportunities: Dict[str, OptionSellOpportunity] = {}
        for symbol in ALL:
            if symbol not in equities:
                continue

            opportunity = self._check_symbol_opportunity(
                symbol=symbol,
                equity=equities[symbol],
            )
            if opportunity is None:
                continue

            opportunities[symbol] = opportunity

        self._slack_opportunities(opportunities=opportunities)

    def _check_symbol_opportunity(
        self,
        symbol: str,
        equity: Equity,
    ) -> Optional[OptionSellOpportunity]:
        redis_key = get_key_weekly_price_pct_chg_sd(symbol=symbol)
        weekly_price_pct_chg_sd_raw = redis.get(name=redis_key)
        if weekly_price_pct_chg_sd_raw is None:
            return None

        weekly_price_pct_chg_sd = json.loads(weekly_price_pct_chg_sd_raw)
        if 'this_week_mean' not in weekly_price_pct_chg_sd:
            print(f'Missing `this_wek_mean`, skipping: {symbol}')
            return None

        this_week_sd = weekly_price_pct_chg_sd['this_week']
        this_week_mean = weekly_price_pct_chg_sd['this_week_mean']
        next_week_sd = weekly_price_pct_chg_sd['next_week']
        next_week_mean = weekly_price_pct_chg_sd['next_week_mean']
        open_price = weekly_price_pct_chg_sd['first_trading_day_open_price']

        pct_chg_so_far = equity.last_price / open_price - 1
        dow = datetime.today().weekday()
        if dow < 3:
            z_score = abs(pct_chg_so_far - this_week_mean) / this_week_sd
        else:
            z_score = abs(pct_chg_so_far - next_week_mean) / next_week_sd

        if z_score <= 1:
            return None

        return OptionSellOpportunity(
            symbol=symbol,
            last_price=equity.last_price,
            pct_chg_so_far=pct_chg_so_far,
            sd_this_week=this_week_sd,
            mean_this_week=this_week_mean,
            sd_next_week=next_week_sd,
            mean_next_week=next_week_mean,
        )

    def _slack_opportunities(self, opportunities: Dict[str, OptionSellOpportunity]) -> None:
        msg = self._construct_slack_msg(opportunities=opportunities)
        slack_client.post_to_option_bot(msg=msg)

    def _construct_slack_msg(self, opportunities: Dict[str, OptionSellOpportunity]) -> str:
        def abs_z_score(opportunity: OptionSellOpportunity) -> float:
            dow = datetime.today().weekday()
            if dow < 3:
                return opportunity.z_score_this_week
            else:
                return opportunity.z_score_next_week

        data = [['symbol', 'last', '%', 'z', 'z next']]
        sorted_by_z_score = sorted(
            opportunities.values(),
            key=abs_z_score,
            reverse=True,
        )
        for opportunity in sorted_by_z_score:
            if opportunity.last_price <= 10:
                continue

            data.append([
                opportunity.symbol,
                '${:,.2f}'.format(opportunity.last_price),
                '{:.2%}'.format(opportunity.pct_chg_so_far),
                '{:.2f}'.format(opportunity.z_score_this_week),
                '{:.2f}'.format(opportunity.z_score_next_week),
            ])
        t = AsciiTable(data)
        t.inner_column_border = True
        t.inner_footing_row_border = False
        t.inner_heading_row_border = True
        t.inner_row_border = False
        t.outer_border = False
        t.justify_columns = {
            0: 'center',
            1: 'right',
            2: 'right',
            3: 'right',
            4: 'right',
        }

        return f'```\n{str(t.table)}```'
